import numpy as np
import matplotlib.pyplot as plt

from enum import Enum

"""
Given a list of coordinates, convert them to screen space.
"""

def compute_pixel_coordinates(
    world_point: np.ndarray,
    transformation_matrix: np.ndarray,
    b: float,
    l: float,
    t: float,
    r: float,
    image_width: int,
    image_height: int
) -> np.ndarray:
    
    # Convert to camera space
    camera_point =  np.append(world_point, 1) @ transformation_matrix

    screen_x = camera_point[0] / -camera_point[2]
    screen_y = camera_point[1] / -camera_point[2]

    normalized_x = (screen_x + r) / (2 * r)
    normalized_y = (screen_y + t) / (2 * t)

    raster_x = np.round(normalized_x * image_width)
    raster_y = np.round(normalized_y * image_height)

    visible = True
    if screen_x < l or screen_x > r or screen_y < b or screen_y > t:
        visible = False


    return np.array([raster_x, raster_y, visible])

FOCALLENGTH = 35
FILMAPERTUREWIDTH = 0.825
FILMAPERTUREHEIGHT = 0.446
INCHTOMM = 25.4
NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

class FitResolutionGate(Enum):
    kFill = 0
    kOverscan = 1

fitFlm = FitResolutionGate.kOverscan

def main():
    points = [
        np.array([   0,  0.5, -1]),
        np.array([ 0.5,    0, -1]),
        np.array([-0.5,    0, -1])
    ]

    # Identity matrix
    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, 1, 0, 0], 
        [0, 0, 1, 0], 
        [0, 0, 50, 1]]
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    print(points)
    print(world_to_camera)

    filmAspectRatio = FILMAPERTUREWIDTH / FILMAPERTUREHEIGHT
    deviceAspectRatio = IMAGE_WIDTH / IMAGE_HEIGHT

    top = ((FILMAPERTUREHEIGHT * INCHTOMM / 2) / FOCALLENGTH) * NEARCLIPPINGPLANE
    right = ((FILMAPERTUREWIDTH * INCHTOMM / 2) / FOCALLENGTH) * NEARCLIPPINGPLANE

    xscale = 1
    yscale = 1

    if fitFlm == FitResolutionGate.kFill:
        if (filmAspectRatio > deviceAspectRatio):
            xscale = deviceAspectRatio / filmAspectRatio
        else:
            yscale = filmAspectRatio / deviceAspectRatio
    elif fitFlm == FitResolutionGate.kOverscan:
            if (filmAspectRatio > deviceAspectRatio):
                yscale = filmAspectRatio / deviceAspectRatio
            else:
                xscale = deviceAspectRatio / filmAspectRatio

    right *= xscale
    top *= yscale

    bottom = -top
    left = -right

    print(f"{bottom=}, {left=}, {top=}, {right=}")

    transformed_points = np.array([compute_pixel_coordinates(p, world_to_camera, bottom, left, top, right, IMAGE_WIDTH, IMAGE_HEIGHT) for p in points])

    print(transformed_points)

    x, y, visible = transformed_points.T
    plt.scatter(x, y)
    plt.ylim((0, IMAGE_HEIGHT))
    plt.xlim((0, IMAGE_WIDTH))
    plt.savefig("fig.png")

if __name__ == "__main__":
    main()