import numpy as np
import matplotlib.pyplot as plt

from enum import Enum


FOCALLENGTH = 20
FILMAPERTUREWIDTH = 0.825
FILMAPERTUREHEIGHT = 0.446
INCHTOMM = 25.4
NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

class FitResolutionGate(Enum):
    kFill = 0
    kOverscan = 1

fitFlm = FitResolutionGate.kOverscan



def compute_pixel_coordinates(
    world_point: np.ndarray,
    transformation_matrix: np.ndarray,
    b: float,
    l: float,
    t: float,
    r: float,
    image_width: int,
    image_height: int
) -> np.ndarray:
    """
    Given a list of coordinates, convert them to screen space.
    """
    
    # Convert to camera space
    camera_point =  np.append(world_point, 1) @ transformation_matrix

    screen_x = camera_point[0] / -camera_point[2]
    screen_y = camera_point[1] / -camera_point[2]
    screen_z = -camera_point[2]

    normalized_x = (screen_x - l) / (r - l)
    normalized_y = (screen_y - b) / (t - b)

    raster_x = np.round(normalized_x * image_width)
    raster_y = np.round(normalized_y * image_height)

    visible = True
    if screen_x < l or screen_x > r or screen_y < b or screen_y > t:
        visible = False


    return np.array([raster_x, raster_y, screen_z, visible])

def edge_function(a: np.ndarray, b: np.ndarray, c: np.ndarray) -> bool:
    return (
          (c[0] - a[0])
        * (b[1] - a[1])
        - (c[1] - a[1])
        * (b[0] - a[0])
    )


def main():
    points = [
        np.array([-0.5, 0, -10]),
        np.array([ 0.5, 0, -10]),
        np.array([ 0.5, 0, -11]),
        np.array([-0.5, 0, -11]),
        np.array([0, 1, -10.5])
    ]

    tris = [
        np.array([0, 2, 1]),
        np.array([0, 3, 2]),
        # np.array([0, 4, 1]),
        np.array([1, 4, 2]),
        np.array([3, 4, 0]),
        np.array([2, 3, 4]),
    ]
    
    phi = 0.44

    # Identity matrix
    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0.5, 10, 10, 1]]
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    print(points)
    print(world_to_camera)

    filmAspectRatio = FILMAPERTUREWIDTH / FILMAPERTUREHEIGHT
    deviceAspectRatio = IMAGE_WIDTH / IMAGE_HEIGHT

    top = ((FILMAPERTUREHEIGHT * INCHTOMM / 2) / FOCALLENGTH) * NEARCLIPPINGPLANE
    right = ((FILMAPERTUREWIDTH * INCHTOMM / 2) / FOCALLENGTH) * NEARCLIPPINGPLANE

    xscale = 1
    yscale = 1

    if fitFlm == FitResolutionGate.kFill:
        if (filmAspectRatio > deviceAspectRatio):
            xscale = deviceAspectRatio / filmAspectRatio
        else:
            yscale = filmAspectRatio / deviceAspectRatio
    elif fitFlm == FitResolutionGate.kOverscan:
            if (filmAspectRatio > deviceAspectRatio):
                yscale = filmAspectRatio / deviceAspectRatio
            else:
                xscale = deviceAspectRatio / filmAspectRatio

    right *= xscale
    top *= yscale

    bottom = -top
    left = -right

    print(f"{bottom=}, {left=}, {top=}, {right=}")

    transformed_points = np.array([compute_pixel_coordinates(p, world_to_camera, bottom, left, top, right, IMAGE_WIDTH, IMAGE_HEIGHT) for p in points])

    print(transformed_points)

    fb = np.zeros(shape=(IMAGE_WIDTH, IMAGE_HEIGHT, 3))
    
    zb = np.full(shape=(IMAGE_WIDTH, IMAGE_HEIGHT), fill_value=FARCLIPINGPLANE)

    colors = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [0, 1, 1]])

    fb = np.zeros(shape=(IMAGE_WIDTH, IMAGE_HEIGHT, 3))
    
    zb = np.full(shape=(IMAGE_WIDTH, IMAGE_HEIGHT), fill_value=FARCLIPINGPLANE, dtype=np.float32)

    for i, tri in enumerate(tris):

        v0 = transformed_points[tri[0]]
        v1 = transformed_points[tri[1]]
        v2 = transformed_points[tri[2]]

        zin0 = 1 / v0[2]
        zin1 = 1 / v1[2]
        zin2 = 1 / v2[2]

        area = edge_function(v0, v1, v2)

        for x in range(IMAGE_WIDTH):
            for y in range(IMAGE_HEIGHT):
                p = np.array([x + 0.5, y + 0.5])


                w0 = edge_function(v1, v2, p)
                w1 = edge_function(v2, v0, p)
                w2 = edge_function(v0, v1, p)

                if w0 >= 0 and w1 >= 0 and w2 >= 0:

                    w0 /= area
                    w1 /= area
                    w2 /= area

                    zin = zin0*w0 + zin1*w1 + zin2*w2

                    z = 1 / zin

                    print(f"{x=}, {y=}, {z=}")
                    print(f"{zb[x][y]=}")
                    

                    if z < zb[x][y]:
                        zb[x][y] = z
                        fb[x][y] = colors[i]

    x, y, z, visible = transformed_points.T
    plt.scatter(x, y)
    plt.ylim((0, IMAGE_HEIGHT))
    plt.xlim((0, IMAGE_WIDTH))
    plt.savefig("points.png")

    plt.figure()
    plt.ylim((0, IMAGE_HEIGHT))
    plt.xlim((0, IMAGE_WIDTH))

    plt.imshow(fb.transpose([1, 0, 2]))
    plt.savefig(f"fig.png")

if __name__ == "__main__":
    main()