import numpy as np
import matplotlib.pyplot as plt
from PIL import Image

NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480


def load_image(infilename: str):
    img = Image.open( infilename )
    img.load()
    data = np.asarray( img, dtype="int32" )
    return data

def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
    S = 1/(np.tan((fov/2)*(np.pi/180)))

    return np.array([
        [S, 0, 0, 0],
        [0, S, 0, 0],
        [0, 0,-f/(f-n),-1],
        [0, 0, -f*n/(f-n), 0]
    ])


def compute_pixel_coordinates(
    world_point: np.ndarray,
    transformation_matrix: np.ndarray,
    projection_matrix: np.ndarray,
    image_width: int,
    image_height: int
) -> np.ndarray:
    """
    Given a list of coordinates, convert them to screen space.
    """
    
    # Convert to camera space
    camera_point =  np.append(world_point, 1) @ transformation_matrix

    print(camera_point)

    projected = camera_point @ projection_matrix

    projected /= projected[3]

    print(f"projected: {projected}")

    raster_x = np.round((projected[0] + 1)/2 * image_width)
    raster_y = np.round((projected[1] + 1)/2 * image_height)

    visible = True
    if projected[0] < -1 or projected[0] > 1 or projected[1] < -1 or projected[1] > 1:
        visible = False


    return np.array([raster_x, raster_y, projected[2], visible])

def rasterize(points, texture_points, textures, tris):

    fb = np.zeros(shape=(IMAGE_WIDTH, IMAGE_HEIGHT, 3))

    colors = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [0, 1, 1]])
    
    zb = np.full(shape=(IMAGE_WIDTH, IMAGE_HEIGHT), fill_value=FARCLIPINGPLANE, dtype=np.float32)

    for tri in tris:

        v0 = points[tri[0]]
        v1 = points[tri[1]]
        v2 = points[tri[2]]

        zin0 = 1 / v0[2]
        zin1 = 1 / v1[2]
        zin2 = 1 / v2[2]

        area = edge_function(v0, v1, v2)

        for x in range(IMAGE_WIDTH):
            for y in range(IMAGE_HEIGHT):
                p = np.array([x + 0.5, y + 0.5])


                w0 = edge_function(v1, v2, p)
                w1 = edge_function(v2, v0, p)
                w2 = edge_function(v0, v1, p)

                if w0 >= 0 and w1 >= 0 and w2 >= 0:

                    w0 /= area
                    w1 /= area
                    w2 /= area

                    zin = zin0*w0 + zin1*w1 + zin2*w2

                    z = 1 / zin


                    if z < zb[x][y]:
                        zb[x][y] = z

                        st0 = texture_points[tri[3]]
                        st1 = texture_points[tri[4]]
                        st2 = texture_points[tri[5]]

                        st_s = st0[0] * w0 + st1[0] * w1 + st2[0] * w2
                        st_t = st0[1] * w0 + st1[1] * w1 + st2[1] * w2
                        st_s *= z
                        st_t *= z

                        texture = np.rot90(m=textures[tri[6]], k=-1, axes=(0, 1))

                        texture_width, texture_height, _ = texture.shape

                        s_int = int(st_s * texture_width)
                        t_int = int(st_t * texture_height)


                        fb[x][y] = texture[s_int][t_int]/256


    return fb

def edge_function(a: np.ndarray, b: np.ndarray, c: np.ndarray) -> bool:
    return (
          (c[0] - a[0])
        * (b[1] - a[1])
        - (c[1] - a[1])
        * (b[0] - a[0])
    )


def main():
    points = [
        np.array([-0.5, 0, -1]),
        np.array([ 0.5, 0, -1]),
        np.array([ 0.5, 0, -2]),
        np.array([-0.5, 0, -2]),
        np.array([0, 1, -1.5])
    ]

    # 3 point indexes, 3 texture point indexes, 1 texture index
    tris = [
        np.array([0, 2, 1, 0, 1, 2, 0]),
        np.array([0, 3, 2, 0, 1, 2, 0]),
        np.array([0, 4, 1, 0, 1, 2, 0]),
        np.array([1, 4, 2, 5, 4, 3, 0]),
        np.array([3, 4, 0, 3, 4, 5, 0]),
        np.array([2, 3, 4, 0, 1, 2, 0]),
    ]

    texture_points = [
        np.array([0.2, 0.35]),
        np.array([0.5, 1]),
        np.array([0.85, 0.35]),
        np.array([0, 0.35]),
        np.array([0.2, 1]),
        np.array([0.4, 0.25]),
        np.array([1, 0.35]),
        np.array([0.8, 1]),
        np.array([0.6, 0.25]),
    ]

    
    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]]
    )

    obama = load_image("obama.ppm")

    print(obama.shape)

    textures = [obama]

    fov = 60

    proj_mat = calc_projection_matrix(fov, NEARCLIPPINGPLANE, FARCLIPINGPLANE)

    print(proj_mat)

    world_to_camera = np.linalg.inv(camera_to_world)

    print(points)
    print(world_to_camera)

    transformed_points = np.array([compute_pixel_coordinates(p, world_to_camera, proj_mat, IMAGE_WIDTH, IMAGE_HEIGHT) for p in points])

    print(transformed_points)
    
    fb = rasterize(transformed_points, texture_points, textures, tris)

    x, y, z, visible = transformed_points.T
    plt.scatter(x, y)
    plt.ylim((0, IMAGE_HEIGHT))
    plt.xlim((0, IMAGE_WIDTH))
    plt.savefig("points.png")

    plt.figure()
    plt.ylim((0, IMAGE_HEIGHT))
    plt.xlim((0, IMAGE_WIDTH))

    plt.imshow(fb.transpose([1, 0, 2]))
    plt.savefig(f"fig.png")

if __name__ == "__main__":
    main()