import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import struct

import logging

import numpy as np


CLK_PERIOD = 5

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.input_queue = Queue()

        self.expected_queue = Queue()
        self.output_queue = Queue()


        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        cocotb.start_soon(self.run_input())
        cocotb.start_soon(self.run_output())

    async def write_input(self, a: np.float32, b: np.float32):
        await self.input_queue.put((a, b))
        self.expected_queue.put_nowait(a + b)

    async def run_input(self):
        while True:
            a, b = await self.input_queue.get()

            self.set_fp32("a", a)
            self.set_fp32("b", b)

            self.dut.i_valid.value = 1

            await RisingEdge(self.dut.clk)
            self.dut.i_valid.value = 0

    async def run_output(self):
        while True:
            await RisingEdge(self.dut.clk)
            if self.dut.o_valid.value == 1:
                self.output_queue.put_nowait(self.get_fp32("z"))

    def set_fp32(self, name, val: np.float32):
        bits, = struct.unpack('<l', val.tobytes())

        sign = bits >> 31
        exponent = (bits >> 23) & 0xff
        mantissa = bits & ((1 << 23) -1)

        getattr(self.dut, f"{name}_sign").value = sign
        getattr(self.dut, f"{name}_exponent").value = exponent
        getattr(self.dut, f"{name}_mantissa").value = mantissa

    def get_fp32(self, name) -> np.float32:
        sign = getattr(self.dut, f"{name}_sign").value
        exponent = getattr(self.dut, f"{name}_exponent").value
        mantissa = getattr(self.dut, f"{name}_mantissa").value

        val = (sign << 31) | (exponent << 23) | (mantissa)

        fp = np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

        return fp

@cocotb.test
async def test_zero(dut):
    tb = TB(dut)

    await RisingEdge(tb.dut.clk)
    await RisingEdge(tb.dut.clk)

    await tb.write_input(np.float32(0), np.float32(4))
    await tb.write_input(np.float32(4), np.float32(0))
    await tb.write_input(np.float32(0), np.float32(0))
    await tb.write_input(np.float32(-0), np.float32(0))

    fail = False

    for _ in range(4):
        dut_val = await tb.output_queue.get()
        expected = tb.expected_queue.get_nowait()

        if dut_val != expected:
            tb.log.info(expected)
            tb.log.info(dut_val)

            fail = True

    assert not fail


N_SAMPLES = 10000

@cocotb.test
async def test_fp32_add(dut):
    tb = TB(dut)

    await RisingEdge(tb.dut.clk)
    await RisingEdge(tb.dut.clk)

    fail = False

    for _ in range(N_SAMPLES):

        a, b = np.random.random_sample(size=2)

        a *= 100
        a -= 50

        b *= 100
        b -= 50

        await tb.write_input(np.float32(a), np.float32(b))

    for _ in range(N_SAMPLES):
        dut_val = await tb.output_queue.get()
        expected = tb.expected_queue.get_nowait()

        if dut_val != expected:
            tb.log.info(expected)
            tb.log.info(dut_val)

            fail = True


    assert not fail
