import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import struct

import logging

import numpy as np


CLK_PERIOD = 5

class TB:
    constant_matrix: np.ndarray

    def __init__(self, dut):
        self.dut = dut

        self.input_queue = Queue()

        self.expected_queue = Queue()
        self.output_queue = Queue()


        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        cocotb.start_soon(self.run_input())
        cocotb.start_soon(self.run_output())

    async def write_constant_matrix(self, m: np.ndarray):
        self.constant_matrix = m

        for i, row in enumerate(m):
            for j, col in enumerate(row):
                self.set_fp32(f"a{i}{j}", np.float32(col))

    async def write_input(self, a: np.ndarray):
        await self.input_queue.put(a)
        self.expected_queue.put_nowait(a @ self.constant_matrix)

    async def run_input(self):
        while True:
            a = await self.input_queue.get()

            for i, val in enumerate(a):
                self.set_fp32(f"v{i}", val)

            self.dut.i_valid.value = 1

            await RisingEdge(self.dut.clk)
            self.dut.i_valid.value = 0

    async def run_output(self):
        while True:
            await RisingEdge(self.dut.clk)
            if self.dut.o_valid.value == 1:
                val = []
                for i in range(4):
                    val.append(self.get_fp32(f"y{i}"))
                self.output_queue.put_nowait(np.array(val))

    def set_fp32(self, name, val: np.float32):
        val = np.float32(val)
        bits, = struct.unpack('<l', val.tobytes())

        sign = bits >> 31
        exponent = (bits >> 23) & 0xff
        mantissa = bits & ((1 << 23) -1)

        getattr(self.dut, f"{name}_sign").value = sign
        getattr(self.dut, f"{name}_exponent").value = exponent
        getattr(self.dut, f"{name}_mantissa").value = mantissa

    def get_fp32(self, name) -> np.float32:
        sign = getattr(self.dut, f"{name}_sign").value
        exponent = getattr(self.dut, f"{name}_exponent").value
        mantissa = getattr(self.dut, f"{name}_mantissa").value

        val = (sign << 31) | (exponent << 23) | (mantissa)

        fp = np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

        return fp
    
@cocotb.test
async def test_fp32_vec4mat44mul(dut):
    tb = TB(dut)

    await RisingEdge(tb.dut.clk)
    await RisingEdge(tb.dut.clk)


    points = [
        np.array([-0.5, 0, -1,   1], dtype=np.float32),
        np.array([ 0.5, 0, -1,   1], dtype=np.float32),
        np.array([ 0.5, 0, -2,   1], dtype=np.float32),
        np.array([-0.5, 0, -2,   1], dtype=np.float32),
        np.array([   0, 1, -1.5, 1], dtype=np.float32)
    ]

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [1, 2, -.4, 1]],
        dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)


    await tb.write_constant_matrix(world_to_camera)

    for point in points:
        await tb.write_input(point)


    fail = False

    for _ in points:
        dut_val = await tb.output_queue.get()
        expected = await tb.expected_queue.get()

        if dut_val.all() != expected.all():

            print(dut_val)
            print(expected)

            fail = True

    assert not fail
