import fp32_pkg::*;

module fp32_add(
    input clk,

    input i_valid,
    
    input fp32_t a,
    input fp32_t b,
    output fp32_t z,

    output logic o_valid
);


logic [2:0] valid_sr;

logic a_is_greater, a_is_greater_next;
logic [7:0] exp_diff, exp_diff_next;

logic [7:0] exponent_sr_next;
logic [7:0] exponent_sr[3];

fp32_t a_sr [3];
fp32_t b_sr [3];

logic [fp32_mantissa_width*2:0] a_shifted, a_shifted_next;
logic [fp32_mantissa_width*2:0] b_shifted, b_shifted_next;

logic [fp32_mantissa_width*2+1:0] mantissa_result, mantissa_result_next;
logic [fp32_mantissa_width*2+1:0] mantissa_remainder;
logic [5:0] leading_zeroes;
logic round_up;

fp32_t result, result_next;

logic sign_adjust, sign_adjust_next;

function bit calculate_sign(bit sign_a, bit sign_b, bit sign_adjust);
    case ({sign_a, sign_b})
        2'b00: return 0;
        2'b01: return sign_adjust;
        2'b10: return ~sign_adjust;
        2'b11: return 1;
    endcase 
endfunction

function logic [5:0] count_leading_zeroes(bit [47:0] val);
    for (bit [5:0] i = 0; i < 47; i++) begin
        if (val[47-i]) begin
            return i;
        end
    end
    return 48;
endfunction

function bit is_zero(fp32_t val);
    return val.exponent == '0 && val.mantissa == '0;
endfunction

assign z = result;

always_ff @(posedge clk) begin
    for (int i = 0; i < 2; i++) begin : sr_update
        a_sr[i+1] <= a_sr[i];
        b_sr[i+1] <= b_sr[i];
        exponent_sr[i+1] <= exponent_sr[i];
    end
    a_sr[0] <= a;
    b_sr[0] <= b;

    exponent_sr[0] <= exponent_sr_next;
    valid_sr <= {valid_sr[1:0], i_valid};
    o_valid <= valid_sr[2];

    a_is_greater <= a_is_greater_next;
    exp_diff <= exp_diff_next;

    a_shifted <= a_shifted_next;
    b_shifted <= b_shifted_next;

    mantissa_result <= mantissa_result_next;
    sign_adjust <= sign_adjust_next;

    result <= result_next;
end

always_comb begin

    // Step 1: caluclate exponent difference
    if (a.exponent > b.exponent) begin
        a_is_greater_next = '1;
        exp_diff_next = a.exponent - b.exponent;
        exponent_sr_next = a.exponent;
    end else begin
        a_is_greater_next = '0;
        exp_diff_next = b.exponent - a.exponent;
        exponent_sr_next = b.exponent;
    end

    // Step 2: shift mantissa of smaller number
    if (a_is_greater) begin
        if (exp_diff > 2*fp32_mantissa_width+1) begin
            b_shifted_next = '0;
        end else begin
            b_shifted_next = {1'b1, b_sr[0].mantissa, fp32_mantissa_width'(0)} >> exp_diff;
        end
        a_shifted_next = {1'b1, a_sr[0].mantissa, fp32_mantissa_width'(0)};
    end else begin
        if (exp_diff > 2*fp32_mantissa_width+1) begin
            a_shifted_next = '0;
        end else begin
            a_shifted_next = {1'b1, a_sr[0].mantissa, fp32_mantissa_width'(0)} >> exp_diff;
        end
        b_shifted_next = {1'b1, b_sr[0].mantissa, fp32_mantissa_width'(0)};
    end

    // Step 3: do the actual add or subtract of mantissas
    case ({a_sr[1].sign, b_sr[1].sign})

        2'b00: begin    // a + b
            mantissa_result_next = a_shifted + b_shifted;
            sign_adjust_next = '0;
        end

        2'b01: begin    // a-b
            if (a_shifted >= b_shifted) begin       // if b > a, do -(b-a)
                mantissa_result_next = a_shifted - b_shifted;
                sign_adjust_next = '0;
            end else begin
                mantissa_result_next = b_shifted - a_shifted;
                sign_adjust_next = '1;
            end
        end

        2'b10: begin    // b-a
            if (b_shifted >= a_shifted) begin       // if a > b, do -(a-b)
                mantissa_result_next = b_shifted - a_shifted;
                sign_adjust_next = '1;
            end else begin
                mantissa_result_next = a_shifted - b_shifted;
                sign_adjust_next = '0;
            end
        end
        
        2'b11: begin    // -a-b  just do -(a+b)
            mantissa_result_next = a_shifted + b_shifted;
            sign_adjust_next = '0;
        end

    endcase

    // Step 4: Normalize mantissa
    // or don't
    leading_zeroes = count_leading_zeroes(mantissa_result);
    mantissa_remainder = mantissa_result << leading_zeroes;
    round_up = &mantissa_remainder[fp32_mantissa_width +: 2] | (mantissa_remainder[fp32_mantissa_width] & |mantissa_remainder[fp32_mantissa_width-1:0]);

    if (is_zero(a_sr[2])) begin
        result_next = b_sr[2];
    end else if (is_zero(b_sr[2])) begin
        result_next = a_sr[2];
    end else begin
        result_next.mantissa = mantissa_result[fp32_mantissa_width-leading_zeroes+1 +: fp32_mantissa_width]+ 23'(round_up);
        result_next.exponent = exponent_sr[2] + 1 - 8'(leading_zeroes);
        result_next.sign = calculate_sign(a_sr[2].sign, b_sr[2].sign, sign_adjust); 
    end
end


endmodule