import fp32_pkg::*;

module fp32_div(
    input clk,
    input i_valid,

    input fp32_t a,
    input fp32_t b,
    output fp32_t z,

    output logic o_valid
);

localparam N_STAGES = fp32_mantissa_width+3;

logic [fp32_exp_width-1:0] exp_sr [N_STAGES+1];
logic [fp32_exp_width-1:0] exp_sr_next;

logic sign_sr[N_STAGES+1];
logic sign_sr_next;


logic valid_sr[N_STAGES+1];

logic [N_STAGES-1:0] quotient_temp [N_STAGES+1];
logic [N_STAGES-1:0] quotient_temp_next [N_STAGES+1];

logic [N_STAGES-1:0] dividend_temp [N_STAGES+1];
logic [N_STAGES-1:0] dividend_temp_next [N_STAGES+1];

logic [N_STAGES-1:0] divisor_sr [N_STAGES+1];
logic [N_STAGES-1:0] divisor_sr_next;

logic result_is_zero_sr [N_STAGES+1];
logic result_is_zero_sr_next;

fp32_t result, result_next;

logic round_up;

assign z = result;


function bit check_bounds(bit[fp32_exp_width-1:0] a, bit[fp32_exp_width-1:0] b);
    automatic bit [fp32_exp_width:0] tmp = b + fp32_exp_bias;
    return ((fp32_exp_width+1)'(a) < tmp);
endfunction


always_ff @(posedge clk) begin
    for (int i = 0; i < N_STAGES; i++) begin
        exp_sr[i+1] <= exp_sr[i];
        divisor_sr[i+1] <= divisor_sr[i];
        valid_sr[i+1] <= valid_sr[i];
        sign_sr[i+1] <= sign_sr[i];
        result_is_zero_sr[i+1] <= result_is_zero_sr[i];
    end

    for (int i = 0; i <= N_STAGES; i++) begin
        quotient_temp[i] <= quotient_temp_next[i];
        dividend_temp[i] <= dividend_temp_next[i];
    end

    exp_sr[0] <= exp_sr_next;
    divisor_sr[0] <= divisor_sr_next;
    valid_sr[0] <= i_valid;
    sign_sr[0] <= a.sign ^ b.sign;
    result_is_zero_sr[0] <= result_is_zero_sr_next;

    result <= result_next;
    o_valid <= valid_sr[N_STAGES];
end


always_comb begin
    // calculate exponent difference and pre-process mantissa.
    // check if divisor or dividend is zero

    // if a < b + 127 and b < a + 127 then we are normal

    case ({check_bounds(a.exponent, b.exponent), check_bounds(b.exponent, a.exponent)})
    2'b00: exp_sr_next = '1;    // this is impossible
    2'b01: exp_sr_next = '1;    // a is much bigger than b, result is infinity
    2'b10: exp_sr_next = '0;    // b is much bigger than a, result is zero.
    2'b11: exp_sr_next = fp32_exp_bias + a.exponent - b.exponent;
    endcase

    result_is_zero_sr_next = a.exponent == '0 && a.mantissa == '0;
    
    divisor_sr_next = {1'b0, 1'b1, b.mantissa, 1'b0};
    dividend_temp_next[0] = {1'b0, 1'b1, a.mantissa, 1'b0};


    for (int i = 0; i < N_STAGES; i++) begin
        if (divisor_sr[i] <= dividend_temp[i]) begin
            dividend_temp_next[i+1] = (dividend_temp[i] - divisor_sr[i]) * 2;
            quotient_temp_next[i+1] = quotient_temp[i] | (1 << (N_STAGES-1-i));
        end else begin
            dividend_temp_next[i+1] = dividend_temp[i] * 2;
            quotient_temp_next[i+1] = quotient_temp[i];
        end
    end

    casez (quotient_temp[N_STAGES][N_STAGES-2 +: 2])
        2'b00: begin
            result_next.exponent = '1;
            result_next.mantissa = 1; // impossible?
            result_next.sign = 0;
        end

        2'b01: begin
            round_up = (quotient_temp[N_STAGES][0] && |dividend_temp[N_STAGES]);
            result_next.exponent = exp_sr[N_STAGES] - 1;
            result_next.mantissa = quotient_temp[N_STAGES][1 +: fp32_mantissa_width] + 23'(round_up);
            result_next.sign = sign_sr[N_STAGES];
        end

        2'b1?: begin
            round_up = quotient_temp[N_STAGES][1] & (quotient_temp[N_STAGES][0] | |dividend_temp[N_STAGES]);
            result_next.exponent = exp_sr[N_STAGES];
            result_next.mantissa = quotient_temp[N_STAGES][2 +: fp32_mantissa_width] + 23'(round_up);
            result_next.sign = sign_sr[N_STAGES];
        end
    endcase

    if (result_is_zero_sr[N_STAGES]) begin
        result_next.exponent = '0;
        result_next.mantissa = '0;
    end
end

endmodule