import fp32_pkg::*;

module fp32_from_int #(
    parameter WIDTH = 32   
)(
    input clk,

    input i_valid,
    
    input logic [WIDTH-1:0] a,
    output fp32_t z,

    output logic o_valid
);

// Converting from fixed point to floating point
// is very simple. The fixed point part becomes
// the mantissa, and the exponent is determined
// by the number of leading zeroes.
//
// for example, 27 in binary is 00000000000000000000000000011011
// that is 11011 with 27 leading zeroes.
//
// The exponent is calcluated with the following formula:
//      WIDTH-1-leading_zeroes+127
// so in this example, its 32-1-27+127 = 131
//
// The mantissa remains the same, except it needs to be bitshifted
// to the left by the mantissa width minus the number of nonzero-bits,
// and the first bit is removed.
//
// In this example, there are 4 nonzero bits (WIDTH-leading_zeroes-1)
// and we have 23 mantissa bits (implied 1). So, we shift 11011 to
// the left by 19 to get 10110000000000000000000
// and our exponent is what we calculated before, 131.
//
// Sign is handled separately. If the number is 2s complement, then we
// convert it to positive first and then change the sign at the end.

fp32_t result, result_next;

logic [1:0] valid_sr;

logic [1:0] is_negative_sr;
logic is_negative_sr_next;

logic [1:0] is_zero_sr;
logic is_zero_sr_next;


logic [WIDTH-1:0] val_positive_d1, val_positive, val_positive_next;

logic [$clog2(WIDTH):0] leading_zeroes, leading_zeroes_next;

assign z = result;

always_ff @(posedge clk) begin
    for (int i = 0; i < 1; i++) begin
        is_negative_sr[i+1] <= is_negative_sr[i];
        is_zero_sr[i+1] <= is_zero_sr[i];
        valid_sr[i+1] <= valid_sr[i];
    end

    leading_zeroes <= leading_zeroes_next;

    is_negative_sr[0] <= is_negative_sr_next;
    is_zero_sr[0] <= is_zero_sr_next;
    valid_sr[0] <= i_valid;

    o_valid <= valid_sr[1];

    val_positive <= val_positive_next;
    val_positive_d1 <= val_positive;

    result <= result_next;
end

function logic [$clog2(WIDTH):0] count_leading_zeroes(bit [WIDTH-1:0] val);
    for (bit [$clog2(WIDTH):0] i = 0; i < WIDTH; i++) begin
        if (val[WIDTH-1-i]) begin
            return i;
        end
    end
    return WIDTH;
endfunction

always_comb begin
    is_zero_sr_next = a == '0;

    if (a[WIDTH-1]) begin
        val_positive_next = ~a + 1;     // convert to positive
        is_negative_sr_next = '1;
    end else begin
        val_positive_next = a;
        is_negative_sr_next = '0;
    end

    leading_zeroes_next = count_leading_zeroes(val_positive);

    if (is_zero_sr[1]) begin
        result_next.sign = '0;
        result_next.mantissa = '0;
        result_next.exponent = '0;
    end else begin
        result_next.sign = is_negative_sr[1];
        result_next.mantissa = fp32_mantissa_width'(val_positive_d1 << fp32_mantissa_width-(WIDTH-leading_zeroes-1));
        result_next.exponent = fp32_exp_bias+8'(WIDTH)-1-8'(leading_zeroes);
    end

end

endmodule