/*

Copyright (c) 2020 Alex Forencich

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/

// Language: Verilog 2001

`resetall
`timescale 1ns / 1ps
`default_nettype none

/*
 * AXI4 lite 1x4 interconnect (wrapper)
 */
module axil_interconnect_wrap_1x4 #
(
    parameter DATA_WIDTH = 32,
    parameter ADDR_WIDTH = 16,
    parameter STRB_WIDTH = (DATA_WIDTH/8),
    parameter M_REGIONS = 1,
    parameter M00_BASE_ADDR = 0,
    parameter M00_ADDR_WIDTH = {M_REGIONS{32'd24}},
    parameter M00_CONNECT_READ = 1'b1,
    parameter M00_CONNECT_WRITE = 1'b1,
    parameter M00_SECURE = 1'b0,
    parameter M01_BASE_ADDR = 0,
    parameter M01_ADDR_WIDTH = {M_REGIONS{32'd24}},
    parameter M01_CONNECT_READ = 1'b1,
    parameter M01_CONNECT_WRITE = 1'b1,
    parameter M01_SECURE = 1'b0,
    parameter M02_BASE_ADDR = 0,
    parameter M02_ADDR_WIDTH = {M_REGIONS{32'd24}},
    parameter M02_CONNECT_READ = 1'b1,
    parameter M02_CONNECT_WRITE = 1'b1,
    parameter M02_SECURE = 1'b0,
    parameter M03_BASE_ADDR = 0,
    parameter M03_ADDR_WIDTH = {M_REGIONS{32'd24}},
    parameter M03_CONNECT_READ = 1'b1,
    parameter M03_CONNECT_WRITE = 1'b1,
    parameter M03_SECURE = 1'b0
)
(
    input  wire                     clk,
    input  wire                     rst,

    /*
     * AXI lite slave interfaces
     */
    axil_intf.SLAVE                 s00_axil,

    /*
     * AXI lite master interfaces
     */
    axil_intf.MASTER                m00_axil,

    axil_intf.MASTER                m01_axil,

    axil_intf.MASTER                m02_axil,

    axil_intf.MASTER                m03_axil
);

localparam S_COUNT = 1;
localparam M_COUNT = 4;

// parameter sizing helpers
function [ADDR_WIDTH*M_REGIONS-1:0] w_a_r(input [ADDR_WIDTH*M_REGIONS-1:0] val);
    w_a_r = val;
endfunction

function [32*M_REGIONS-1:0] w_32_r(input [32*M_REGIONS-1:0] val);
    w_32_r = val;
endfunction

function [S_COUNT-1:0] w_s(input [S_COUNT-1:0] val);
    w_s = val;
endfunction

function w_1(input val);
    w_1 = val;
endfunction

axil_interconnect #(
    .S_COUNT(S_COUNT),
    .M_COUNT(M_COUNT),
    .DATA_WIDTH(DATA_WIDTH),
    .ADDR_WIDTH(ADDR_WIDTH),
    .STRB_WIDTH(STRB_WIDTH),
    .M_REGIONS(M_REGIONS),
    .M_BASE_ADDR({ w_a_r(M03_BASE_ADDR), w_a_r(M02_BASE_ADDR), w_a_r(M01_BASE_ADDR), w_a_r(M00_BASE_ADDR) }),
    .M_ADDR_WIDTH({ w_32_r(M03_ADDR_WIDTH), w_32_r(M02_ADDR_WIDTH), w_32_r(M01_ADDR_WIDTH), w_32_r(M00_ADDR_WIDTH) }),
    .M_CONNECT_READ({ w_s(M03_CONNECT_READ), w_s(M02_CONNECT_READ), w_s(M01_CONNECT_READ), w_s(M00_CONNECT_READ) }),
    .M_CONNECT_WRITE({ w_s(M03_CONNECT_WRITE), w_s(M02_CONNECT_WRITE), w_s(M01_CONNECT_WRITE), w_s(M00_CONNECT_WRITE) }),
    .M_SECURE({ w_1(M03_SECURE), w_1(M02_SECURE), w_1(M01_SECURE), w_1(M00_SECURE) })
)
axil_interconnect_inst (
    .clk(clk),
    .rst(rst),
    .s_axil_awaddr({ s00_axil.awaddr }),
    .s_axil_awprot({ s00_axil.awprot }),
    .s_axil_awvalid({ s00_axil.awvalid }),
    .s_axil_awready({ s00_axil.awready }),
    .s_axil_wdata({ s00_axil.wdata }),
    .s_axil_wstrb({ s00_axil.wstrb }),
    .s_axil_wvalid({ s00_axil.wvalid }),
    .s_axil_wready({ s00_axil.wready }),
    .s_axil_bresp({ s00_axil.bresp }),
    .s_axil_bvalid({ s00_axil.bvalid }),
    .s_axil_bready({ s00_axil.bready }),
    .s_axil_araddr({ s00_axil.araddr }),
    .s_axil_arprot({ s00_axil.arprot }),
    .s_axil_arvalid({ s00_axil.arvalid }),
    .s_axil_arready({ s00_axil.arready }),
    .s_axil_rdata({ s00_axil.rdata }),
    .s_axil_rresp({ s00_axil.rresp }),
    .s_axil_rvalid({ s00_axil.rvalid }),
    .s_axil_rready({ s00_axil.rready }),
    .m_axil_awaddr({ m03_axil.awaddr, m02_axil.awaddr, m01_axil.awaddr, m00_axil.awaddr }),
    .m_axil_awprot({ m03_axil.awprot, m02_axil.awprot, m01_axil.awprot, m00_axil.awprot }),
    .m_axil_awvalid({ m03_axil.awvalid, m02_axil.awvalid, m01_axil.awvalid, m00_axil.awvalid }),
    .m_axil_awready({ m03_axil.awready, m02_axil.awready, m01_axil.awready, m00_axil.awready }),
    .m_axil_wdata({ m03_axil.wdata, m02_axil.wdata, m01_axil.wdata, m00_axil.wdata }),
    .m_axil_wstrb({ m03_axil.wstrb, m02_axil.wstrb, m01_axil.wstrb, m00_axil.wstrb }),
    .m_axil_wvalid({ m03_axil.wvalid, m02_axil.wvalid, m01_axil.wvalid, m00_axil.wvalid }),
    .m_axil_wready({ m03_axil.wready, m02_axil.wready, m01_axil.wready, m00_axil.wready }),
    .m_axil_bresp({ m03_axil.bresp, m02_axil.bresp, m01_axil.bresp, m00_axil.bresp }),
    .m_axil_bvalid({ m03_axil.bvalid, m02_axil.bvalid, m01_axil.bvalid, m00_axil.bvalid }),
    .m_axil_bready({ m03_axil.bready, m02_axil.bready, m01_axil.bready, m00_axil.bready }),
    .m_axil_araddr({ m03_axil.araddr, m02_axil.araddr, m01_axil.araddr, m00_axil.araddr }),
    .m_axil_arprot({ m03_axil.arprot, m02_axil.arprot, m01_axil.arprot, m00_axil.arprot }),
    .m_axil_arvalid({ m03_axil.arvalid, m02_axil.arvalid, m01_axil.arvalid, m00_axil.arvalid }),
    .m_axil_arready({ m03_axil.arready, m02_axil.arready, m01_axil.arready, m00_axil.arready }),
    .m_axil_rdata({ m03_axil.rdata, m02_axil.rdata, m01_axil.rdata, m00_axil.rdata }),
    .m_axil_rresp({ m03_axil.rresp, m02_axil.rresp, m01_axil.rresp, m00_axil.rresp }),
    .m_axil_rvalid({ m03_axil.rvalid, m02_axil.rvalid, m01_axil.rvalid, m00_axil.rvalid }),
    .m_axil_rready({ m03_axil.rready, m02_axil.rready, m01_axil.rready, m00_axil.rready })
);

endmodule

`resetall
