module gpu_top (
    input  wire             sys_clk_p,
    input  wire             sys_clk_n,

    input  wire             rst_n,

    input  wire             pcie_exp_clkp,
    input  wire             pcie_exp_clkn,

    output wire [1 : 0]     pci_exp_txp,
    output wire [1 : 0]     pci_exp_txn,
    input  wire [1 : 0]     pci_exp_rxp,
    input  wire [1 : 0]     pci_exp_rxn,

    output wire [3:0]       led,


    inout [31:0]            ddr3_dq,
    inout [3:0]             ddr3_dqs_n,
    inout [3:0]             ddr3_dqs_p,

    output [14:0]           ddr3_addr,
    output [2:0]            ddr3_ba,
    output                  ddr3_ras_n,
    output                  ddr3_cas_n,
    output                  ddr3_we_n,
    output                  ddr3_reset_n,
    output [0:0]            ddr3_ck_p,
    output [0:0]            ddr3_ck_n,
    output [0:0]            ddr3_cke,
    output [0:0]            ddr3_cs_n,
    output [3:0]            ddr3_dm,
    output [0:0]            ddr3_odt
);

logic pcie_refclk;

logic clk_100;
logic clk_200;
logic clk_400;

logic link_up;
logic init_calib_complete;

assign led[0] = ~link_up;
assign led[1] = ~init_calib_complete;

logic rst_100;

assign rst_100 = ~rst_n;

// always @(posedge clk_100) begin
//     rst_100 <= ~rst_n;          // dumb single flop
// end


wire sys_clk_w;
wire sys_clk;

`ifndef VERILATOR
IBUFDS sys_clk_ibufgds
(
    .O  (sys_clk_w),
    .I  (sys_clk_p),
    .IB (sys_clk_n)
);

BUFG bufg_clk (
   .O(sys_clk), // 1-bit output: Clock output
   .I(sys_clk_w)  // 1-bit input: Clock input
);
`else
    assign sys_clk_w = sys_clk_p;
    assign sys_clk = sys_clk_w;
`endif

axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) m_axil_pcie();
axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) m_axil_vertex_gen();
axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) m_axil_primitive_processor();
axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) m_axil_rasterizer();

axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) m_axil_ctrl_vertex_gen();
axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) m_axil_ctrl_primitive_processor();
axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) m_axil_ctrl_rasterizer();


axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) s_axil_ddr();

axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) s_axil_cfg_pcie();
axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) s_axil_cfg_vertex_gen();
axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) s_axil_cfg_primitive_processor();
axil_intf #(.AXIL_DATA_WIDTH(32), .AXIL_ADDR_WIDTH(32)) s_axil_cfg_rasterizer();

`ifndef VERILATOR
IBUFDS_GTE2 #(
    .CLKRCV_TRST("TRUE"),
    .CLKCM_CFG("TRUE"),
    .CLKSWING_CFG(2'b11)
) pcie_ibuf (
    .I                      (pcie_exp_clkp),
    .IB                     (pcie_exp_clkn),
    .CEB                    ('0),
    .O                      (pcie_refclk),
    .ODIV2                  ()
);
`else
    assign pcie_refclk = pcie_exp_clkp;
`endif

clk_wiz_0 u_clk_wiz_0 (
    .clk_in1                (sys_clk),

    .clk_out1               (clk_100),
    .clk_out2               (clk_200),
    .clk_out3               (clk_400)
);

pcie_wrapper u_pcie_wrapper (
    .clk_pcie               (pcie_refclk),
    .rstn                   (rst_n),

    .rst_100                (rst_100),
    .clk_100                (clk_100),

    .link_up                (link_up),

    .m_data_axil            (m_axil_pcie),
    .s_ctrl_axil            (s_axil_cfg_pcie),

    .pci_exp_txp            (pci_exp_txp),
    .pci_exp_txn            (pci_exp_txn),
    .pci_exp_rxp            (pci_exp_rxp),
    .pci_exp_rxn            (pci_exp_rxn)
);

vertex_generator u_vertex_generator (
    .clk                    (clk_100),
    .rst                    (rst_100),

    .m_data_axil            (m_axil_vertex_gen),
    .m_ctrl_axil            (m_axil_ctrl_vertex_gen),
    .s_ctrl_axil            (s_axil_cfg_vertex_gen)
);

interconnect_top u_interconnect_top (
    .clk                    (clk_100),
    .rst                    (rst_100),

    .s_axil_pcie                (m_axil_pcie),
    .s_axil_vertex_gen          (m_axil_vertex_gen),
    .s_axil_primitive_processor (m_axil_primitive_processor),
    .s_axil_rasterizer          (m_axil_rasterizer),

    .s_axil_ctrl_vertex_gen             (m_axil_ctrl_vertex_gen),
    .s_axil_ctrl_primitive_processor    (m_axil_ctrl_primitive_processor),
    .s_axil_ctrl_rasterizer             (m_axil_ctrl_rasterizer),

    .m_axil_ddr             (s_axil_ddr),

    .m_axil_cfg_pcie        (s_axil_cfg_pcie),
    .m_axil_cfg_vertex_gen  (s_axil_cfg_vertex_gen),
    .m_axil_cfg_primitive_processor (s_axil_cfg_primitive_processor),
    .m_axil_cfg_rasterizer  (s_axil_cfg_rasterizer)
);


mig_wrapper u_mig_wrapper (
    // Memory interface ports
    .ddr3_addr              (ddr3_addr),
    .ddr3_ba                (ddr3_ba),
    .ddr3_cas_n             (ddr3_cas_n),
    .ddr3_ck_n              (ddr3_ck_n),
    .ddr3_ck_p              (ddr3_ck_p),
    .ddr3_cke               (ddr3_cke),
    .ddr3_ras_n             (ddr3_ras_n),
    .ddr3_reset_n           (ddr3_reset_n),
    .ddr3_we_n              (ddr3_we_n),
    .ddr3_dq                (ddr3_dq),
    .ddr3_dqs_n             (ddr3_dqs_n),
    .ddr3_dqs_p             (ddr3_dqs_p),
      
	.ddr3_cs_n              (ddr3_cs_n),
    .ddr3_dm                (ddr3_dm),
    .ddr3_odt               (ddr3_odt),

    .init_calib_complete    (init_calib_complete),

    .sys_clk                (sys_clk),
    .clk_100                (clk_100),

    .aresetn                (rst_n),

    .s_axil                 (s_axil_ddr)
);

endmodule