import cocotb

from cocotb.clock import Clock

from cocotbext.axi import AxiLiteMaster, AxiLiteBus, AxiLiteRam
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import numpy as np
import struct


import logging

CLK_100_PERIOD = 10
CLK_125_PERIOD = 8
CLK_200_PERIOD = 5

class TB:
    NEARCLIPPINGPLANE = 0.1
    FARCLIPINGPLANE = 1000

    IMAGE_WIDTH = 640
    IMAGE_HEIGHT = 480


    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(self.dut.u_clk_wiz_0.clk_out1, CLK_100_PERIOD, units="ns").start())
        cocotb.start_soon(Clock(self.dut.sys_clk_p, CLK_200_PERIOD, units="ns").start())
        cocotb.start_soon(Clock(self.dut.pcie_exp_clkp, CLK_100_PERIOD, units="ns").start())
        cocotb.start_soon(Clock(self.dut.u_pcie_wrapper.u_axi_pcie_0.axi_aclk_out, CLK_125_PERIOD, units="ns").start())
        cocotb.start_soon(Clock(self.dut.u_mig_wrapper.u_ddr3.ui_clk, CLK_100_PERIOD, units="ns").start())

        self.pcie_axil_master = AxiLiteMaster(AxiLiteBus.from_prefix(dut.u_pcie_wrapper.u_axi_pcie_0, "m_axi"), self.dut.u_pcie_wrapper.u_axi_pcie_0.axi_aclk_out, self.dut.rst_100)
        self.ddr_mem_axil = AxiLiteRam(AxiLiteBus.from_prefix(self.dut.u_mig_wrapper.u_ddr3, "s_axi"), self.dut.u_mig_wrapper.u_ddr3.ui_clk, self.dut.u_mig_wrapper.u_ddr3.ui_clk_sync_rst, size=2**30)



    async def cycle_reset(self):
        await self._cycle_reset(self.dut.rst_n, self.dut.clk_100, inverted=True)
        await self._cycle_reset(self.dut.u_mig_wrapper.u_ddr3.ui_clk_sync_rst, self.dut.u_mig_wrapper.u_ddr3.ui_clk)

    async def _cycle_reset(self, rst, clk, inverted=False):
        if inverted:
            rst.setimmediatevalue(1)
            await RisingEdge(clk)
            await RisingEdge(clk)
            rst.value = 0
            await RisingEdge(clk)
            await RisingEdge(clk)
            rst.value = 1
            await RisingEdge(clk)
            await RisingEdge(clk)
        else:
            rst.setimmediatevalue(0)
            await RisingEdge(clk)
            await RisingEdge(clk)
            rst.value = 1
            await RisingEdge(clk)
            await RisingEdge(clk)
            rst.value = 0
            await RisingEdge(clk)
            await RisingEdge(clk)

    @staticmethod
    def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
        S = 1/(np.tan((fov/2)*(np.pi/180)))

        return np.array([
            [S, 0, 0, 0],
            [0, S, 0, 0],
            [0, 0,-f/(f-n),-1],
            [0, 0, -f*n/(f-n), 0]
        ], dtype=np.float32)
    
    @staticmethod
    def fp32_to_int(val: np.float32):
        bits, = struct.unpack('<L', val.tobytes())
        return bits

    @staticmethod
    def int_to_fp32(val: int):
        return np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]
    
    @staticmethod
    def compute_pixel_coordinates(
        world_point: np.ndarray,
        transformation_matrix: np.ndarray,
        projection_matrix: np.ndarray,
        image_width: int,
        image_height: int
    ) -> np.ndarray:
        """
        Given a list of coordinates, convert them to screen space.
        """
        
        # Convert to camera space
        camera_point =  np.append(world_point, np.float32(1)) @ transformation_matrix

        # print(camera_point)

        projected = camera_point @ projection_matrix

        projected /= projected[3]

        print(f"projected: {projected}")

        raster_x = (projected[0] + 1)/2 * image_width
        raster_y = (projected[1] + 1)/2 * image_height

        visible = True
        if projected[0] < -1 or projected[0] > 1 or projected[1] < -1 or projected[1] > 1:
            visible = False


        return np.array([raster_x, raster_y, projected[2], visible])