from gpu_testbench import TB

import cocotb

from decimal import Decimal
from cocotb.triggers import Timer

import numpy as np

@cocotb.test
async def test_sanity(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    points = [
        np.array([-0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -2], dtype=np.float32),
        np.array([-0.5, 0, -2], dtype=np.float32),
        np.array([   0, 1, -1.5], dtype=np.float32)
    ]

    vertex_descriptor_addr = 0x20

    world_to_camera_matrix_ptr = 0x40
    projection_matrix_ptr = 0x80
    screen_dim_ptr = 0xc0
    vertex_source_ptr = 0x100
    vertex_dest_ptr = 0x200
    vertex_buffer_length = 5

    await tb.pcie_axil_master.write_dword(vertex_descriptor_addr+0x0, world_to_camera_matrix_ptr)
    await tb.pcie_axil_master.write_dword(vertex_descriptor_addr+0x4, projection_matrix_ptr)
    await tb.pcie_axil_master.write_dword(vertex_descriptor_addr+0x8, screen_dim_ptr)
    await tb.pcie_axil_master.write_dword(vertex_descriptor_addr+0xc, vertex_source_ptr)
    await tb.pcie_axil_master.write_dword(vertex_descriptor_addr+0x10, vertex_dest_ptr)
    await tb.pcie_axil_master.write_dword(vertex_descriptor_addr+0x14, vertex_buffer_length)

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]], dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    tb.log.info(world_to_camera)

    for i, row in enumerate(world_to_camera):
        for j, col in enumerate(row):
            await tb.pcie_axil_master.write_dword(world_to_camera_matrix_ptr + i*16 + j*4, tb.fp32_to_int(np.float32(col)))

    fov = 60
    proj_mat = tb.calc_projection_matrix(fov, tb.NEARCLIPPINGPLANE, tb.FARCLIPINGPLANE)

    tb.log.info(proj_mat)

    for i, row in enumerate(proj_mat):
        for j, col in enumerate(row):
            await tb.pcie_axil_master.write_dword(projection_matrix_ptr + i*16 + j*4, tb.fp32_to_int(np.float32(col)))

    await tb.pcie_axil_master.write_dword(screen_dim_ptr, tb.fp32_to_int(np.float32(tb.IMAGE_WIDTH)))
    await tb.pcie_axil_master.write_dword(screen_dim_ptr + 4, tb.fp32_to_int(np.float32(tb.IMAGE_HEIGHT)))

    for i, point in enumerate(points):
        for j, val in enumerate(point):
            await tb.pcie_axil_master.write_dword(vertex_source_ptr + i*12 + j*4, tb.fp32_to_int(np.float32(val)))
    
    await tb.pcie_axil_master.write_word(0x80010004, vertex_descriptor_addr)

    transformed_points = np.array([tb.compute_pixel_coordinates(p, world_to_camera, proj_mat, tb.IMAGE_WIDTH, tb.IMAGE_HEIGHT) for p in points])

    # wait for done, read reg?
    await Timer(Decimal(40000), "ns")

    fail = False

    for i, point in enumerate(transformed_points):
        dut_val = np.array([tb.int_to_fp32(await tb.pcie_axil_master.read_dword(vertex_dest_ptr + i*16 + j*4)) for j in range(3)])
        
        for a, b in zip(dut_val, point[0:3]):

            if a != b:
                print(f"{i} dut: {dut_val}")
                print(f"{i} exp: {point}")

                fail = True
                break

    tb.log.info(transformed_points)

    assert not fail

    for i, point in enumerate(points):
        for j, val in enumerate(point):
            await tb.pcie_axil_master.write_dword(vertex_source_ptr + i*12 + j*4, tb.fp32_to_int(np.float32(val)))
    