import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

from cocotbext.axi import AxiLiteMaster, AxiLiteBus, AxiLiteRam

import struct

import logging

import numpy as np
import matplotlib.pyplot as plt

CLK_PERIOD = 5

NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

def compute_pixel_coordinates(
    world_point: np.ndarray,
    transformation_matrix: np.ndarray,
    projection_matrix: np.ndarray,
    image_width: int,
    image_height: int
) -> np.ndarray:
    """
    Given a list of coordinates, convert them to screen space.
    """
    
    # Convert to camera space
    camera_point =  np.append(world_point, np.float32(1)) @ transformation_matrix

    # print(camera_point)

    projected = camera_point @ projection_matrix

    projected /= projected[3]

    print(f"projected: {projected}")

    raster_x = (projected[0] + 1)/2 * image_width
    raster_y = (projected[1] + 1)/2 * image_height

    visible = True
    if projected[0] < -1 or projected[0] > 1 or projected[1] < -1 or projected[1] > 1:
        visible = False


    return np.array([raster_x, raster_y, projected[2], visible])

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        self.m_ctrl_axil = AxiLiteRam(AxiLiteBus.from_prefix(dut.m_ctrl_axil, ""), dut.clk, dut.rst, size=2**32)
        self.m_data_axil = AxiLiteRam(AxiLiteBus.from_prefix(dut.m_data_axil, ""), dut.clk, dut.rst, size=2**32)
        self.s_ctrl_axil = AxiLiteMaster(AxiLiteBus.from_prefix(dut.s_ctrl_axil, ""), dut.clk, dut.rst)


    async def cycle_reset(self):
        await self._cycle_reset(self.dut.rst, self.dut.clk)


    async def _cycle_reset(self, rst, clk):
        rst.setimmediatevalue(0)
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 1
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 0
        await RisingEdge(clk)
        await RisingEdge(clk)

def fp32_to_int(val: np.float32):
    bits, = struct.unpack('<L', val.tobytes())
    return bits

def int_to_fp32(val: int):
    return np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]


def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
    S = 1/(np.tan((fov/2)*(np.pi/180)))

    return np.array([
        [S, 0, 0, 0],
        [0, S, 0, 0],
        [0, 0,-f/(f-n),-1],
        [0, 0, -f*n/(f-n), 0]
    ], dtype=np.float32)

def compare_matrix(A, B):
    fail = False
    for A_row, B_row in zip(A, B):
        for A_col, B_col in zip (A_row, B_row):
            if A_col != B_col:
                print(f"{A_col} {B_col}")
                fail = True
    return fail

@cocotb.test
async def test_sanity(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    points = [
        np.array([-0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -2], dtype=np.float32),
        np.array([-0.5, 0, -2], dtype=np.float32),
        np.array([   0, 1, -1.5], dtype=np.float32)
    ]

    vertex_descriptor_addr = 0x20

    world_to_camera_matrix_ptr = 0x40
    projection_matrix_ptr = 0x80
    screen_dim_ptr = 0xc0
    vertex_source_ptr = 0x100
    vertex_dest_ptr = 0x200
    vertex_buffer_length = 5

    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x0, world_to_camera_matrix_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x4, projection_matrix_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x8, screen_dim_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0xc, vertex_source_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x10, vertex_dest_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x14, vertex_buffer_length)

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]], dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    tb.log.info(world_to_camera)

    for i, row in enumerate(world_to_camera):
        for j, col in enumerate(row):
            tb.m_ctrl_axil.write_dword(world_to_camera_matrix_ptr + i*16 + j*4, fp32_to_int(np.float32(col)))

    fov = 60
    proj_mat = calc_projection_matrix(fov, NEARCLIPPINGPLANE, FARCLIPINGPLANE)

    tb.log.info(proj_mat)

    for i, row in enumerate(proj_mat):
        for j, col in enumerate(row):
            tb.m_ctrl_axil.write_dword(projection_matrix_ptr + i*16 + j*4, fp32_to_int(np.float32(col)))

    tb.m_ctrl_axil.write_dword(screen_dim_ptr, fp32_to_int(np.float32(IMAGE_WIDTH)))
    tb.m_ctrl_axil.write_dword(screen_dim_ptr + 4, fp32_to_int(np.float32(IMAGE_HEIGHT)))

    await tb.s_ctrl_axil.write_word(0x4, vertex_descriptor_addr)

    for i, point in enumerate(points):
        for j, val in enumerate(point):
            tb.m_data_axil.write_dword(vertex_source_ptr + i*12 + j*4, fp32_to_int(np.float32(val)))


    await RisingEdge(tb.dut.u_dut.done)

    transformed_points = np.array([compute_pixel_coordinates(p, world_to_camera, proj_mat, IMAGE_WIDTH, IMAGE_HEIGHT) for p in points])

    print(transformed_points)

    fail = False

    for i, point in enumerate(transformed_points):
        dut_val = np.array([int_to_fp32(tb.m_data_axil.read_dword(vertex_dest_ptr + i*16 + j*4)) for j in range(3)])
        
        for a, b in zip(dut_val, point):

            if a != b:
                print(f"{i} dut: {dut_val}")
                print(f"{i} exp: {point}")

                fail = True
                break

    x, y, z, visible = transformed_points.T
    plt.scatter(x, y)
    plt.ylim((0, IMAGE_HEIGHT))
    plt.xlim((0, IMAGE_WIDTH))
    plt.savefig("points.png")

    assert not fail