import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

from cocotbext.axi import AxiLiteMaster, AxiLiteBus, AxiLiteRam

import struct

import logging

import numpy as np

CLK_PERIOD = 5

NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        self.m_ctrl_axil = AxiLiteRam(AxiLiteBus.from_prefix(dut.m_ctrl_axil, ""), dut.clk, dut.rst, size=2**32)
        self.s_ctrl_axil = AxiLiteMaster(AxiLiteBus.from_prefix(dut.s_ctrl_axil, ""), dut.clk, dut.rst)


    async def cycle_reset(self):
        await self._cycle_reset(self.dut.rst, self.dut.clk)


    async def _cycle_reset(self, rst, clk):
        rst.setimmediatevalue(0)
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 1
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 0
        await RisingEdge(clk)
        await RisingEdge(clk)

    def read_matrix(self, name: str):
        m = np.zeros(shape=(4,4), dtype=np.float32)

        for i, row in enumerate(m):
            for j, col in enumerate(row):
                m[i][j] = self.get_fp32(f"{name}{i}{j}")

        return m

    def get_fp32(self, name) -> np.float32:
        sign = getattr(self.dut, f"{name}_sign").value
        exponent = getattr(self.dut, f"{name}_exponent").value
        mantissa = getattr(self.dut, f"{name}_mantissa").value

        val = (sign << 31) | (exponent << 23) | (mantissa)

        fp = np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

        return fp


def fp32_to_int(val: np.float32):
    bits, = struct.unpack('<L', val.tobytes())
    return bits

def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
    S = 1/(np.tan((fov/2)*(np.pi/180)))

    return np.array([
        [S, 0, 0, 0],
        [0, S, 0, 0],
        [0, 0,-f/(f-n),-1],
        [0, 0, -f*n/(f-n), 0]
    ], dtype=np.float32)

def compare_matrix(A, B):
    fail = False
    for A_row, B_row in zip(A, B):
        for A_col, B_col in zip (A_row, B_row):
            if A_col != B_col:
                print(f"{A_col} {B_col}")
                fail = True
    return fail

@cocotb.test
async def test_sanity(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    vertex_descriptor_addr = 0x20

    world_to_camera_matrix_ptr = 0x40
    projection_matrix_ptr = 0x80
    screen_dim_ptr = 0xc0
    vertex_source_ptr = 0xc8
    vertex_dest_ptr = 0xcc
    vertex_buffer_length = 0xd0

    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x0, world_to_camera_matrix_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x4, projection_matrix_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x8, screen_dim_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0xc, vertex_source_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x10, vertex_dest_ptr)
    tb.m_ctrl_axil.write_dword(vertex_descriptor_addr+0x14, vertex_buffer_length)

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]], dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    tb.log.info(world_to_camera)

    for i, row in enumerate(world_to_camera):
        for j, col in enumerate(row):
            tb.m_ctrl_axil.write_dword(world_to_camera_matrix_ptr + i*16 + j*4, fp32_to_int(np.float32(col)))

    fov = 60
    proj_mat = calc_projection_matrix(fov, NEARCLIPPINGPLANE, FARCLIPINGPLANE)

    tb.log.info(proj_mat)

    for i, row in enumerate(proj_mat):
        for j, col in enumerate(row):
            tb.m_ctrl_axil.write_dword(projection_matrix_ptr + i*16 + j*4, fp32_to_int(np.float32(col)))

    tb.m_ctrl_axil.write_dword(screen_dim_ptr, fp32_to_int(np.float32(IMAGE_WIDTH)))
    tb.m_ctrl_axil.write_dword(screen_dim_ptr + 4, fp32_to_int(np.float32(IMAGE_HEIGHT)))


    await tb.s_ctrl_axil.write_word(0x4, vertex_descriptor_addr)

    await RisingEdge(tb.dut.start)


    fail = False

    if compare_matrix(tb.read_matrix("world_to_camera_matrix"), world_to_camera):
        fail = True

    if compare_matrix(tb.read_matrix("projection_matrix"), proj_mat):
        fail = True

    dut_screen_x = tb.get_fp32("screen_x")
    if dut_screen_x != np.float32(IMAGE_WIDTH):
        tb.log.error(f"{dut_screen_x} {np.float32(IMAGE_WIDTH)}")
        fail = True

    dut_screen_y = tb.get_fp32("screen_y")
    if dut_screen_y != np.float32(IMAGE_HEIGHT):
        tb.log.error(f"{dut_screen_y} {np.float32(IMAGE_HEIGHT)}")
        fail = True

    if tb.dut.vertex_buffer_source.value != vertex_source_ptr:
        tb.log.error(f"{tb.dut.vertex_buffer_source.value:x} {vertex_source_ptr:x}")
        fail = True

    if tb.dut.vertex_buffer_dest.value != vertex_dest_ptr:
        tb.log.error(f"{tb.dut.vertex_buffer_dest.value:x} {vertex_dest_ptr:x}")
        fail = True

    if tb.dut.vertex_buffer_len.value != vertex_buffer_length:
        tb.log.error(f"{tb.dut.vertex_buffer_len.value:x} {vertex_buffer_length:x}")
        fail = True

    await RisingEdge(tb.dut.clk)
    await RisingEdge(tb.dut.clk)

    if tb.dut.u_dut.state.value != 7:
        tb.log.error(f"Not in state 7: {tb.dut.u_dut.state.value}")

    tb.dut.done.value = 1
    await RisingEdge(tb.dut.clk)
    tb.dut.done.value = 0
    await RisingEdge(tb.dut.clk)


    if tb.dut.u_dut.state.value != 0:
        tb.log.error(f"Not in state 0: {tb.dut.u_dut.state.value}")

    assert not fail