import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import struct

import logging

import numpy as np


CLK_PERIOD = 5

NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

def compute_pixel_coordinates(
    world_point: np.ndarray,
    transformation_matrix: np.ndarray,
    projection_matrix: np.ndarray,
    image_width: int,
    image_height: int
) -> np.ndarray:
    """
    Given a list of coordinates, convert them to screen space.
    """
    
    # Convert to camera space
    camera_point =  np.append(world_point, np.float32(1)) @ transformation_matrix

    projected = camera_point @ projection_matrix

    projected /= projected[3]

    raster_x = (projected[0] + np.float32(1))/np.float32(2) * np.float32(image_width)
    raster_y = (projected[1] + np.float32(1))/np.float32(2) * np.float32(image_height)

    visible = True
    if projected[0] < -1 or projected[0] > 1 or projected[1] < -1 or projected[1] > 1:
        visible = False


    return np.array([raster_x, raster_y, projected[2], visible])

def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
    S = 1/(np.tan((fov/2)*(np.pi/180)))

    return np.array([
        [S, 0, 0, 0],
        [0, S, 0, 0],
        [0, 0,-f/(f-n),-1],
        [0, 0, -f*n/(f-n), 0]
    ], dtype=np.float32)

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.input_queue = Queue()

        self.expected_queue = Queue()
        self.output_queue = Queue()


        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        screen_x: np.float32
        screen_y: np.float32

        transformation_matrix: np.ndarray
        projection_matrix: np.ndarray

        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        cocotb.start_soon(self.run_input())
        cocotb.start_soon(self.run_output())

    def set_raster_dims(self, x: np.float32, y: np.float32):
        self.set_fp32("screen_x", x)
        self.set_fp32("screen_y", y)
        
        self.screen_x = x
        self.screen_y = y

    async def write_constant_matrix(self, name: str, m: np.ndarray):
        for i, row in enumerate(m):
            for j, col in enumerate(row):
                self.set_fp32(f"{name}{i}{j}", np.float32(col))

    async def write_input(self, a: np.ndarray):
        await self.input_queue.put(a)

        self.expected_queue.put_nowait(
            compute_pixel_coordinates(
                a, 
                self.transformation_matrix, 
                self.projection_matrix,
                IMAGE_WIDTH,
                IMAGE_HEIGHT
            )
        )

    async def run_input(self):
        while True:
            a = await self.input_queue.get()

            for i, val in enumerate(a):
                self.set_fp32(f"world{i}", val)

            self.dut.i_valid.value = 1

            await RisingEdge(self.dut.clk)
            self.dut.i_valid.value = 0

    async def run_output(self):
        while True:
            await RisingEdge(self.dut.clk)
            if self.dut.o_valid.value == 1:
                val = []
                for i in range(3):
                    val.append(self.get_fp32(f"raster{i}"))
                val.append(self.dut.o_visible.value == 1)

                self.output_queue.put_nowait(np.array(val))

    def set_fp32(self, name, val: np.float32):
        val = np.float32(val)
        bits, = struct.unpack('<l', val.tobytes())

        sign = bits >> 31
        exponent = (bits >> 23) & 0xff
        mantissa = bits & ((1 << 23) -1)

        getattr(self.dut, f"{name}_sign").value = sign
        getattr(self.dut, f"{name}_exponent").value = exponent
        getattr(self.dut, f"{name}_mantissa").value = mantissa

    def get_fp32(self, name) -> np.float32:
        sign = getattr(self.dut, f"{name}_sign").value
        exponent = getattr(self.dut, f"{name}_exponent").value
        mantissa = getattr(self.dut, f"{name}_mantissa").value

        val = (sign << 31) | (exponent << 23) | (mantissa)

        fp = np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

        return fp


@cocotb.test
async def test_sanity(dut):
    tb = TB(dut)

    await RisingEdge(tb.dut.clk)
    await RisingEdge(tb.dut.clk)

    tb.set_raster_dims(np.float32(IMAGE_WIDTH), np.float32(IMAGE_HEIGHT))

    points = [
        np.array([-0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -2], dtype=np.float32),
        np.array([-0.5, 0, -2], dtype=np.float32),
        np.array([   0, 1, -1.5], dtype=np.float32)
    ]

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]],
        dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)


    fov = 60
    proj_mat = calc_projection_matrix(fov, NEARCLIPPINGPLANE, FARCLIPINGPLANE)


    await tb.write_constant_matrix("world_to_camera_matrix", world_to_camera)
    tb.transformation_matrix = world_to_camera

    await tb.write_constant_matrix("projection_matrix", proj_mat)
    tb.projection_matrix = proj_mat

    for point in points:
        await tb.write_input(point)

    fail = False

    for _ in points:
        dut_val = await tb.output_queue.get()
        expected = await tb.expected_queue.get()


        for a, b in zip(dut_val, expected):

            if a != b:
                print(f"dut: {dut_val}")
                print(f"exp: {expected}")

                fail = True
                break

    assert not fail