import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import struct

import logging

import numpy as np


CLK_PERIOD = 5

NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.input_queue = Queue()

        self.expected_queue = Queue()
        self.output_queue = Queue()


        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        screen_x: np.float32
        screen_y: np.float32

        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        cocotb.start_soon(self.run_input())
        cocotb.start_soon(self.run_output())

    def set_raster_dims(self, x: np.float32, y: np.float32):
        self.set_fp32("screen_x", x)
        self.set_fp32("screen_y", y)
        
        self.screen_x = x
        self.screen_y = y

    async def write_input(self, a: np.ndarray, visible: bool):
        await self.input_queue.put((a, visible))

        raster_x = (a[0] + 1)/2 * self.screen_x
        raster_y = (a[1] + 1)/2 * self.screen_y


        self.expected_queue.put_nowait(np.array([raster_x, raster_y, a[2], visible]))

    async def run_input(self):
        while True:
            a, visible = await self.input_queue.get()

            for i, val in enumerate(a[0:3]):
                self.set_fp32(f"v{i}", val)

            self.dut.i_visible.value = visible

            self.dut.i_valid.value = 1

            await RisingEdge(self.dut.clk)
            self.dut.i_valid.value = 0

    async def run_output(self):
        while True:
            await RisingEdge(self.dut.clk)
            if self.dut.o_valid.value == 1:
                val = []
                for i in range(3):
                    val.append(self.get_fp32(f"y{i}"))

                visible = self.dut.o_visible.value == 1
                self.output_queue.put_nowait(np.array([val[0], val[1], val[2], visible]))

    def set_fp32(self, name, val: np.float32):
        val = np.float32(val)
        bits, = struct.unpack('<l', val.tobytes())

        sign = bits >> 31
        exponent = (bits >> 23) & 0xff
        mantissa = bits & ((1 << 23) -1)

        getattr(self.dut, f"{name}_sign").value = sign
        getattr(self.dut, f"{name}_exponent").value = exponent
        getattr(self.dut, f"{name}_mantissa").value = mantissa

    def get_fp32(self, name) -> np.float32:
        sign = getattr(self.dut, f"{name}_sign").value
        exponent = getattr(self.dut, f"{name}_exponent").value
        mantissa = getattr(self.dut, f"{name}_mantissa").value

        val = (sign << 31) | (exponent << 23) | (mantissa)

        fp = np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

        return fp
    
def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
    S = 1/(np.tan((fov/2)*(np.pi/180)))

    return np.array([
        [S, 0, 0, 0],
        [0, S, 0, 0],
        [0, 0,-f/(f-n),-1],
        [0, 0, -f*n/(f-n), 0]
    ], dtype=np.float32)

@cocotb.test
async def test_vertex_generator_raster_coordinate_gen(dut):
    tb = TB(dut)

    await RisingEdge(tb.dut.clk)
    await RisingEdge(tb.dut.clk)

    points = [
        np.array([-0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -2], dtype=np.float32),
        np.array([-0.5, 0, -2], dtype=np.float32),
        np.array([0, 1, -1.5], dtype=np.float32)
    ]

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]],
        dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    fov = 60

    proj_mat = calc_projection_matrix(fov, NEARCLIPPINGPLANE, FARCLIPINGPLANE)

    tb.set_raster_dims(np.float32(IMAGE_WIDTH), np.float32(IMAGE_HEIGHT))


    for point in points:
        camera_point =  np.append(point, np.float32(1)) @ world_to_camera

        projected = camera_point @ proj_mat

        projected /= projected[3]

        visible = True
        if projected[0] < -1 or projected[0] > 1 or projected[1] < -1 or projected[1] > 1:
            visible = False

        await tb.write_input(projected, visible)

    fail = False

    for _ in points:
        dut_val = await tb.output_queue.get()
        expected = await tb.expected_queue.get()

        for a, b in zip(dut_val, expected):

            if a != b:
                print(dut_val)
                print(expected)

                fail = True
                break

    assert not fail
