import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

from cocotbext.axi import AxiLiteMaster, AxiLiteBus, AxiLiteRam

import struct

import logging

import numpy as np

CLK_PERIOD = 5



class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        self.output_queue = Queue()


        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        self.m_axil_read = AxiLiteRam(AxiLiteBus.from_prefix(dut.m_axil_read, ""), dut.clk, dut.rst, size=2**32)

        cocotb.start_soon(self.run_output())


    async def cycle_reset(self):
        await self._cycle_reset(self.dut.rst, self.dut.clk)


    async def _cycle_reset(self, rst, clk):
        rst.setimmediatevalue(0)
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 1
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 0
        await RisingEdge(clk)
        await RisingEdge(clk)

    async def run_output(self):
        while True:
            await RisingEdge(self.dut.clk)
            if self.dut.valid.value == 1:
                val = []
                for i in range(3):
                    val.append(self.get_fp32(f"vertex{i}"))

                self.output_queue.put_nowait(np.array(val))


    def get_fp32(self, name) -> np.float32:
        sign = getattr(self.dut, f"{name}_sign").value
        exponent = getattr(self.dut, f"{name}_exponent").value
        mantissa = getattr(self.dut, f"{name}_mantissa").value

        val = (sign << 31) | (exponent << 23) | (mantissa)

        fp = np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

        return fp

def fp32_to_int(val: np.float32):
    bits, = struct.unpack('<L', val.tobytes())
    return bits

@cocotb.test
async def test_sanity(dut):
    tb = TB(dut)

    await tb.cycle_reset()


    points = [
        np.array([-0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -2], dtype=np.float32),
        np.array([-0.5, 0, -2], dtype=np.float32),
        np.array([   0, 1, -1.5], dtype=np.float32)
    ]

    source_addr = 0x20

    for i, point in enumerate(points):
        for j, val in enumerate(point):
            tb.m_axil_read.write_dword(source_addr + i*12 + j*4, fp32_to_int(np.float32(val)))

    tb.dut.ready.value = 1

    tb.dut.source_addr.value = source_addr
    tb.dut.len.value = 3*len(points)

    await RisingEdge(tb.dut.clk)
    tb.dut.start.value = 1
    await RisingEdge(tb.dut.clk)
    tb.dut.start.value = 0

    fail = False

    for point in points:
        dut_val = await tb.output_queue.get()


        for a, b in zip(dut_val, point):

            if a != b:
                print(f"dut: {dut_val}")
                print(f"exp: {point}")

                fail = True
                break

    assert not fail