import cocotb

from cocotb.queue import Queue

from decimal import Decimal

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

from cocotbext.axi import AxiLiteMaster, AxiLiteBus, AxiLiteRam

import struct

import logging

import numpy as np

CLK_PERIOD = 5



class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        self.input_queue = Queue()


        cocotb.start_soon(Clock(self.dut.clk, CLK_PERIOD, units="ns").start())

        self.m_axil_write = AxiLiteRam(AxiLiteBus.from_prefix(dut.m_axil_write, ""), dut.clk, dut.rst, size=2**32)

        cocotb.start_soon(self.run_input())


    async def cycle_reset(self):
        await self._cycle_reset(self.dut.rst, self.dut.clk)


    async def _cycle_reset(self, rst, clk):
        rst.setimmediatevalue(0)
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 1
        await RisingEdge(clk)
        await RisingEdge(clk)
        rst.value = 0
        await RisingEdge(clk)
        await RisingEdge(clk)

    async def write_input(self, a: np.ndarray):
        await self.input_queue.put(a)

    async def run_input(self):
        while True:
            a = await self.input_queue.get()

            for i in range(3):
                self.set_fp32(f"vertex{i}", a[i])
            self.dut.visible.value = 1

            self.dut.valid.value = 1
            await RisingEdge(self.dut.clk)
            while not self.dut.ready.value:
                await RisingEdge(self.dut.clk)
            self.dut.valid.value = 0


    def set_fp32(self, name, val: np.float32):
        val = np.float32(val)
        bits, = struct.unpack('<l', val.tobytes())

        sign = bits >> 31
        exponent = (bits >> 23) & 0xff
        mantissa = bits & ((1 << 23) -1)

        getattr(self.dut, f"{name}_sign").value = sign
        getattr(self.dut, f"{name}_exponent").value = exponent
        getattr(self.dut, f"{name}_mantissa").value = mantissa

def fp32_to_int(val: np.float32):
    bits, = struct.unpack('<L', val.tobytes())
    return bits

def int_to_fp32(val: int):
    return np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

@cocotb.test
async def test_sanity(dut):
    tb = TB(dut)

    await tb.cycle_reset()


    points = [
        np.array([-0.5, 0, -1, 1], dtype=np.float32),
        np.array([ 0.5, 0, -1, 1], dtype=np.float32),
        np.array([ 0.5, 0, -2, 1], dtype=np.float32),
        np.array([-0.5, 0, -2, 1], dtype=np.float32),
        np.array([   0, 1, -1.5, 1], dtype=np.float32),
    ]


    dest_addr = 0x120

    for point in points:
        await tb.write_input(point)


    tb.dut.dest_addr.value = dest_addr
    tb.dut.len.value = len(points)

    await RisingEdge(tb.dut.clk)
    tb.dut.start.value = 1
    await RisingEdge(tb.dut.clk)
    tb.dut.start.value = 0

    await RisingEdge(tb.dut.done)

    fail = False

    for i, point in enumerate(points):
        dut_val = np.array([int_to_fp32(tb.m_axil_write.read_dword(dest_addr + i*16 + j*4)) for j in range(3)])
        
        for a, b in zip(dut_val, point[0:3]):

            if a != b:
                print(f"{i} dut: {dut_val}")
                print(f"{i} exp: {point}")

                fail = True
                break

    assert not fail