module vertex_generator(
    input clk,
    input rst,

    axil_intf.MASTER    m_data_axil,    // main datapath

    axil_intf.MASTER    m_ctrl_axil,    // control data, like matrix/screen coords
    axil_intf.SLAVE     s_ctrl_axil     // control registers/stats
);


axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) m_axil_write();
axil_intf #(.AXIL_DATA_WIDTH(128), .AXIL_ADDR_WIDTH(32)) m_axil_read();

assign m_axil_write.awready = m_data_axil.awready;
assign m_data_axil.awvalid = m_axil_write.awvalid;
assign m_data_axil.awaddr = m_axil_write.awaddr;
assign m_data_axil.awprot = m_axil_write.awprot;
assign m_axil_write.wready = m_data_axil.wready;
assign m_data_axil.wvalid = m_axil_write.wvalid;
assign m_data_axil.wdata = m_axil_write.wdata;
assign m_data_axil.wstrb = m_axil_write.wstrb;
assign m_data_axil.bready = '1;

assign m_axil_read.arready = m_data_axil.arready;
assign m_data_axil.arvalid = m_axil_read.arvalid;
assign m_data_axil.araddr = m_axil_read.araddr;
assign m_data_axil.arprot = m_axil_read.arprot;
assign m_data_axil.rready = m_axil_read.rready;
assign m_axil_read.rvalid = m_data_axil.rvalid;
assign m_axil_read.rdata = m_data_axil.rdata;
assign m_axil_read.rresp = m_data_axil.rresp;

logic world_point_valid;
fp32_t world_point[3];

fp32_t world_to_camera_matrix[4][4];
fp32_t projection_matrix[4][4];
fp32_t screen_dim [2];

fp32_t raster_points[3];
logic raster_points_visible;
logic raster_points_valid;

fp32_t fifo_raster_points[3];
logic fifo_raster_points_visible;
logic fifo_raster_points_valid;

logic [31:0] vertex_buffer_source;
logic [31:0] vertex_buffer_dest;
logic [31:0] vertex_buffer_len;

logic start;

logic read_done, write_done;

logic done;

assign done = read_done & write_done;

logic ready;

vertex_generator_control u_control (
    .clk                    (clk),
    .rst                    (rst),

    .m_ctrl_axil            (m_ctrl_axil),
    .s_ctrl_axil            (s_ctrl_axil),

    .world_to_camera_matrix (world_to_camera_matrix),
    .projection_matrix      (projection_matrix),
    .screen_dim             (screen_dim),

    .vertex_buffer_source   (vertex_buffer_source),
    .vertex_buffer_dest     (vertex_buffer_dest),
    .vertex_buffer_len      (vertex_buffer_len),

    .start                  (start),
    .done                   (done)
);

vertex_generator_read_fsm u_read_fsm(
    .clk                    (clk),
    .rst                    (rst),

    .source_addr            (vertex_buffer_source),
    .len                    (vertex_buffer_len),
    .start                  (start),
    .done                   (read_done),

    .valid                  (world_point_valid),
    .ready                  (ready),
    .vertex                 (world_point),

    .m_axil_read            (m_axil_read)
);

`ifdef DEBUG_ILA
ila_2 world_point_ila (
	.clk(clk), // input wire clk


	.probe0(world_point[0]), // input wire [31:0]  probe0  
	.probe1(world_point[1]), // input wire [31:0]  probe1 
	.probe2(world_point[2]), // input wire [31:0]  probe2 
	.probe3(world_point_valid), // input wire [0:0]  probe3 
	.probe4(ready) // input wire [0:0]  probe4
);
`endif 

vertex_generator_datapath u_datapath (
    .clk                    (clk),
//  .rst                    (rst),

    .i_valid                (world_point_valid),
    .world_point            (world_point),
    .world_to_camera_matrix (world_to_camera_matrix),
    .projection_matrix      (projection_matrix),
    .screen_dim             (screen_dim),

    .raster_points          (raster_points),
    .o_visible              (raster_points_visible),
    .o_valid                (raster_points_valid)
);

axis_fifo #(
    .DEPTH                  (128),
    .DATA_WIDTH             (97),
    .KEEP_ENABLE            ('0),
    .LAST_ENABLE            ('0),
    .USER_ENABLE            ('0)
) u_output_fifo (
    .clk                    (clk),
    .rst                    (rst),

    .s_axis_tdata           ({raster_points[0], raster_points[1], raster_points[2], raster_points_visible}),
    .s_axis_tkeep           ('1),
    .s_axis_tvalid          (raster_points_valid),
    .s_axis_tready          (),
    .s_axis_tlast           ('0),
    .s_axis_tid             ('0),
    .s_axis_tdest           ('0),
    .s_axis_tuser           ('0),

    .m_axis_tdata           ({fifo_raster_points[0], fifo_raster_points[1], fifo_raster_points[2], fifo_raster_points_visible}),
    .m_axis_tkeep           (),
    .m_axis_tvalid          (fifo_raster_points_valid),
    .m_axis_tready          (ready),
    .m_axis_tlast           (),
    .m_axis_tid             (),
    .m_axis_tdest           (),
    .m_axis_tuser           (),

    .pause_req              ('0),
    .pause_ack              (),

    .status_depth           (),
    .status_depth_commit    (),
    .status_overflow        (),
    .status_bad_frame       (),
    .status_good_frame      ()
);

`ifdef DEBUG_ILA
ila_2 raster_point_ila (
	.clk(clk), // input wire clk


	.probe0(fifo_raster_points[0]), // input wire [31:0]  probe0  
	.probe1(fifo_raster_points[1]), // input wire [31:0]  probe1 
	.probe2(fifo_raster_points[2]), // input wire [31:0]  probe2 
	.probe3(fifo_raster_points_valid), // input wire [0:0]  probe3 
	.probe4(ready) // input wire [0:0]  probe4
);

ila_0 ddr_axil_ila (
	.clk(clk), // input wire clk


	.probe0(    m_data_axil.wready), // input wire [0:0] probe0  
	.probe1(    m_data_axil.awaddr), // input wire [31:0]  probe1 
	.probe2(    m_data_axil.bresp), // input wire [1:0]  probe2 
	.probe3(    m_data_axil.bvalid), // input wire [0:0]  probe3 
	.probe4(    m_data_axil.bready), // input wire [0:0]  probe4 
	.probe5(    m_data_axil.araddr), // input wire [31:0]  probe5 
	.probe6(    m_data_axil.rready), // input wire [0:0]  probe6 
	.probe7(    m_data_axil.wvalid), // input wire [0:0]  probe7 
	.probe8(    m_data_axil.arvalid), // input wire [0:0]  probe8 
	.probe9(    m_data_axil.arready), // input wire [0:0]  probe9 
	.probe10(   m_data_axil.rdata), // input wire [127:0]  probe10 
	.probe11(   m_data_axil.awvalid), // input wire [0:0]  probe11 
	.probe12(   m_data_axil.awready), // input wire [0:0]  probe12 
	.probe13(   m_data_axil.rresp), // input wire [1:0]  probe13 
	.probe14(   m_data_axil.wdata), // input wire [127:0]  probe14 
	.probe15(   m_data_axil.wstrb), // input wire [15:0]  probe15 
	.probe16(   m_data_axil.rvalid), // input wire [0:0]  probe16 
	.probe17(   m_data_axil.arprot), // input wire [2:0]  probe17  
	.probe18(   m_data_axil.awprot) // input wire [2:0]  probe18
);


`endif

vertex_generator_write_fsm u_write_fsm(
    .clk                    (clk),
    .rst                    (rst),

    .dest_addr              (vertex_buffer_dest),
    .len                    (vertex_buffer_len),
    .start                  (start),
    .done                   (write_done),

    .valid                  (fifo_raster_points_valid),
    .ready                  (ready),
    .visible                (fifo_raster_points_visible),
    .vertex                 (fifo_raster_points),

    .m_axil_write           (m_axil_write)
);

endmodule