import fp32_pkg::*;

module vertex_generator_control(
    input   clk,
    input   rst,

    axil_intf.MASTER    m_ctrl_axil,    // control data, like matrix/screen coords
    axil_intf.SLAVE     s_ctrl_axil,    // control registers/stats,

    output fp32_t world_to_camera_matrix[4][4],
    output fp32_t projection_matrix[4][4],

    output fp32_t screen_dim [2],

    output logic [31:0] vertex_buffer_source,
    output logic [31:0] vertex_buffer_dest,
    output logic [31:0] vertex_buffer_len,

    output logic start,
    input logic done
);

vertex_generator_regs_pkg::vertex_generator_regs__out_t hwif_out;

vertex_generator_regs u_regs(
    .clk            (clk),
    .rst            (rst),

    .s_axil_awready (s_ctrl_axil.awready),
    .s_axil_awvalid (s_ctrl_axil.awvalid),
    .s_axil_awaddr  (s_ctrl_axil.awaddr[2:0]),
    .s_axil_awprot  (s_ctrl_axil.awprot),
    .s_axil_wready  (s_ctrl_axil.wready),
    .s_axil_wvalid  (s_ctrl_axil.wvalid),
    .s_axil_wdata   (s_ctrl_axil.wdata),
    .s_axil_wstrb   (s_ctrl_axil.wstrb),
    .s_axil_bready  (s_ctrl_axil.bready),
    .s_axil_bvalid  (s_ctrl_axil.bvalid),
    .s_axil_bresp   (s_ctrl_axil.bresp),
    .s_axil_arready (s_ctrl_axil.arready),
    .s_axil_arvalid (s_ctrl_axil.arvalid),
    .s_axil_araddr  (s_ctrl_axil.araddr[2:0]),
    .s_axil_arprot  (s_ctrl_axil.arprot),
    .s_axil_rready  (s_ctrl_axil.rready),
    .s_axil_rvalid  (s_ctrl_axil.rvalid),
    .s_axil_rdata   (s_ctrl_axil.rdata),
    .s_axil_rresp   (s_ctrl_axil.rresp),

    .hwif_out       (hwif_out)
);


fp32_t r_world_to_camera_matrix[4][4];
fp32_t r_world_to_camera_matrix_next[4][4];

fp32_t r_projection_matrix[4][4];
fp32_t r_projection_matrix_next[4][4];

fp32_t r_screen_dim[2];
fp32_t r_screen_dim_next[2];

logic [31:0] r_vertex_buffer_source, r_vertex_buffer_source_next;
logic [31:0] r_vertex_buffer_dest, r_vertex_buffer_dest_next;
logic [31:0] r_vertex_buffer_len, r_vertex_buffer_len_next;

assign world_to_camera_matrix = r_world_to_camera_matrix;
assign projection_matrix = r_projection_matrix;
assign screen_dim = r_screen_dim;

logic [31:0] descriptor_pointer, descriptor_pointer_next;

logic [31:0] descriptor[6];
logic [31:0] descriptor_next[6];

assign vertex_buffer_source = descriptor[3];
assign vertex_buffer_dest = descriptor[4];
assign vertex_buffer_len = descriptor[5];

logic [31:0] counter, counter_next;

logic saw_arready, saw_arready_next;

enum logic [2:0] {
    IDLE,           // Idle state
    SAVE_POINTER,   // Wait one cycle for data in hwif_out to be valid
    READ_DESCRIPTOR,
    READ_WORLD_TO_CAMERA,
    READ_PROJECTION_MATRIX,
    READ_SCREEN_DIMS,
    WAIT,
    BUSY
} state, state_next;

always @(posedge clk) begin
    if (rst) begin
        state <= IDLE;
    end else begin
        state <= state_next;
        counter <= counter_next;


        r_world_to_camera_matrix <= r_world_to_camera_matrix_next;
        r_projection_matrix <= r_projection_matrix_next;
        r_screen_dim <= r_screen_dim_next;

        descriptor_pointer <= descriptor_pointer_next;
        descriptor <= descriptor_next;

        saw_arready <= saw_arready_next;
    end
end

always_comb begin
    state_next = state;
    counter_next = counter;

    r_world_to_camera_matrix_next = r_world_to_camera_matrix;
    r_projection_matrix_next = r_projection_matrix;
    r_screen_dim_next = r_screen_dim;

    descriptor_pointer_next = descriptor_pointer;

    descriptor_next = descriptor;

    saw_arready_next = saw_arready;

    m_ctrl_axil.araddr = '0;
    m_ctrl_axil.arvalid = '0;
    m_ctrl_axil.rready = '0;

    start = '0;


    case (state)
        IDLE: begin
            if (hwif_out.vertex_descriptor.ptr.swmod) begin
                state_next = SAVE_POINTER;
            end

            counter_next = '0;
        end

        SAVE_POINTER: begin
            descriptor_pointer_next = hwif_out.vertex_descriptor.ptr.value;
            state_next = READ_DESCRIPTOR;
        end

        READ_DESCRIPTOR: begin
            if (counter < 6) begin
                m_ctrl_axil.rready = '1;

                m_ctrl_axil.araddr = (counter * 4) + descriptor_pointer;

                m_ctrl_axil.arvalid = !saw_arready;

                if (m_ctrl_axil.arready) begin
                    saw_arready_next = '1;
                end

                if (m_ctrl_axil.rvalid) begin
                    descriptor_next[counter] = m_ctrl_axil.rdata;
                    counter_next = counter + 1;
                    saw_arready_next = '0;
                end

            end else begin
                state_next = READ_WORLD_TO_CAMERA;
                counter_next = '0;
            end
        end 

        READ_WORLD_TO_CAMERA: begin
            m_ctrl_axil.rready = '1;

            m_ctrl_axil.araddr = (counter * 4) + descriptor[0];

            m_ctrl_axil.arvalid = !saw_arready;

            if (m_ctrl_axil.arready) begin
                saw_arready_next = '1;
            end

            if (m_ctrl_axil.rvalid) begin
                r_world_to_camera_matrix_next[counter / 4][counter % 4] = fp32_t'(m_ctrl_axil.rdata);
                counter_next = counter + 1;
                saw_arready_next = '0;

                if (counter >= 15) begin
                    state_next = READ_PROJECTION_MATRIX;
                    counter_next = '0;
                end
            end
        end

        READ_PROJECTION_MATRIX: begin
            m_ctrl_axil.rready = '1;

            m_ctrl_axil.araddr = (counter * 4) + descriptor[1];

            m_ctrl_axil.arvalid = !saw_arready;

            if (m_ctrl_axil.arready) begin
                saw_arready_next = '1;
            end

            if (m_ctrl_axil.rvalid) begin
                r_projection_matrix_next[counter / 4][counter % 4] = fp32_t'(m_ctrl_axil.rdata);
                counter_next = counter + 1;
                saw_arready_next = '0;

                if (counter >= 15) begin
                    state_next = READ_SCREEN_DIMS;
                    counter_next = '0;
                end
            end
        end

        READ_SCREEN_DIMS: begin
            m_ctrl_axil.rready = '1;

            m_ctrl_axil.araddr = (counter * 4) + descriptor[2];

            m_ctrl_axil.arvalid = !saw_arready;

            if (m_ctrl_axil.arready) begin
                saw_arready_next = '1;
            end

            if (m_ctrl_axil.rvalid) begin
                r_screen_dim_next[counter] = fp32_t'(m_ctrl_axil.rdata);
                counter_next = counter + 1;
                saw_arready_next = '0;

                if (counter >= 1) begin
                    state_next = WAIT;
                    counter_next = '0;
                end
            end
        end

        WAIT: begin
            state_next = BUSY;
            start = '1;
        end

        BUSY: begin
            if (done) begin
                state_next = IDLE;
            end
        end

        default: begin
            state_next = IDLE;
        end

    endcase
end

endmodule