from pypcie import Device
import numpy as np
import struct

import time

NEARCLIPPINGPLANE = 0.1
FARCLIPINGPLANE = 1000

IMAGE_WIDTH = 640
IMAGE_HEIGHT = 480

@staticmethod
def calc_projection_matrix(fov: float, n: float, f: float) -> np.ndarray:
    S = 1/(np.tan((fov/2)*(np.pi/180)))

    return np.array([
        [S, 0, 0, 0],
        [0, S, 0, 0],
        [0, 0,-f/(f-n),-1],
        [0, 0, -f*n/(f-n), 0]
    ], dtype=np.float32)

@staticmethod
def fp32_to_int(val: np.float32):
    bits, = struct.unpack('<L', val.tobytes())
    return bits

@staticmethod
def int_to_fp32(val: int):
    return np.frombuffer(struct.pack('<L', val), dtype=np.float32)[0]

def main():
    d = Device("0000:01:00.0")

    mem_bar = d.bar[0]
    ctrl_bar = d.bar[1]

    points = [
        np.array([-0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -1], dtype=np.float32),
        np.array([ 0.5, 0, -2], dtype=np.float32),
        np.array([-0.5, 0, -2], dtype=np.float32),
        np.array([   0, 1, -1.5], dtype=np.float32)
    ]

    vertex_descriptor_addr = 0x20

    world_to_camera_matrix_ptr = 0x40
    projection_matrix_ptr = 0x80
    screen_dim_ptr = 0xc0
    vertex_source_ptr = 0x100
    vertex_dest_ptr = 0x200
    vertex_buffer_length = 5

    mem_bar.write(vertex_descriptor_addr+0x0, world_to_camera_matrix_ptr)
    mem_bar.write(vertex_descriptor_addr+0x4, projection_matrix_ptr)
    mem_bar.write(vertex_descriptor_addr+0x8, screen_dim_ptr)
    mem_bar.write(vertex_descriptor_addr+0xc, vertex_source_ptr)
    mem_bar.write(vertex_descriptor_addr+0x10, vertex_dest_ptr)
    mem_bar.write(vertex_descriptor_addr+0x14, vertex_buffer_length)

    phi = 0.8

    camera_to_world = np.array(
       [[1, 0, 0, 0], 
        [0, np.cos(phi), -np.sin(phi), 0], 
        [0, np.sin(phi), np.cos(phi), 0], 
        [0, 2, -.4, 1]], dtype=np.float32
    )

    world_to_camera = np.linalg.inv(camera_to_world)

    print(world_to_camera)

    for i, row in enumerate(world_to_camera):
        for j, col in enumerate(row):
            mem_bar.write(world_to_camera_matrix_ptr + i*16 + j*4, fp32_to_int(np.float32(col)))

    fov = 60
    proj_mat = calc_projection_matrix(fov, NEARCLIPPINGPLANE, FARCLIPINGPLANE)

    for i, row in enumerate(proj_mat):
        for j, col in enumerate(row):
            mem_bar.write(projection_matrix_ptr + i*16 + j*4, fp32_to_int(np.float32(col)))

    mem_bar.write(screen_dim_ptr, fp32_to_int(np.float32(IMAGE_WIDTH)))
    mem_bar.write(screen_dim_ptr + 4, fp32_to_int(np.float32(IMAGE_HEIGHT)))

    for i, point in enumerate(points):
        for j, val in enumerate(point):
            mem_bar.write(vertex_source_ptr + i*12 + j*4, fp32_to_int(np.float32(val)))
    

    ctrl_bar.write(0x10004, vertex_descriptor_addr)

    time.sleep(1)

    for i, point in enumerate(points):
        dut_val = np.array([int_to_fp32(mem_bar.read(vertex_dest_ptr + i*16 + j*4)) for j in range(3)])
        
        print(f"{i}, {dut_val}")

if __name__ == "__main__":
    main()